STATUSES = {
        200: "200 OK",
        201: "201 Created",
        202: "202 Accepted",
        204: "204 No Content",
        206: "206 Partial Content",
        207: "207 Multi-Status",
        208: "208 Already Reported",
        226: "226 IM Used",
        300: "300 Multiple Choices",
        301: "301 Moved Permanently",
        302: "302 Found",
        303: "303 See Other",
        304: "304 Not Modified",
        305: "305 Use Proxy",
        307: "307 Temporary Redirect",
        308: "308 Permanent Redirect",
        400: "400 Bad Request",
        401: "401 Unauthorized",
        402: "402 Payment Required",
        403: "403 Forbidden",
        404: "404 Not Found",
        405: "405 Method Not Allowed",
        406: "406 Not Acceptable",
        407: "407 Proxy Authentication Required",
        408: "408 Request Timeout",
        409: "409 Conflict",
        410: "410 Gone",
        411: "411 Length Required",
        412: "412 Precondition Failed",
        413: "413 Payload Too Large",
        414: "414 URI Too Long",
        415: "415 Unsupported Media Type",
        416: "416 Range Not Satisfiable",
        417: "417 Expectation Failed",
        421: "421 Misdirected Request",
        422: "422 Unprocessable Entity",
        423: "423 Locked",
        424: "424 Failed Dependency",
        425: "425 Too Early",
        426: "426 Upgrade Required",
        428: "428 Precondition Required",
        429: "429 Too Many Requests",
        431: "431 Request Header Fields Too Large",
        451: "451 Unavailable For Legal Reasons",
        500: "500 Internal Server Error",
        501: "501 Not Implemented",
        502: "502 Bad Gateway",
        503: "503 Service Unavailable",
        504: "504 Gateway Timeout",
        505: "505 HTTP Version Not Supported",
        506: "506 Variant Also Negotiates",
        507: "507 Insufficient Storage",
        508: "508 Loop Detected",
        510: "510 Not Extended",
        577: "577 Unknown Status",
}



#!/usr/bin/env python3
import re
import traceback
import json
import logging

from . import utils
from .data import STATUSES
from . import errors


class Handler(object):

    request = None
    response = None

    def options(self):
        """Default implementation of OPTIONS method.

        Checks which methods are implemented, and sets the Allow header
        accordingly.
        """
        # Reference: https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods
        http_methods = [
            'get',
            'head',
            'post',
            'put',
            'delete',
            'connect',
            'options',
            'trace',
            'patch'
        ]

        supported_methods = []

        for method in http_methods:
            if method in dir(self):
                supported_methods.append(method.upper())

        self.response.set_header('Allow', ', '.join(supported_methods))
        return {}


class Request(object):

    def __init__(self, path, method, environ):
        self._headers = None
        self._args = None
        self._json_args = None
        self._cookies = None
        self._body = None
        self._data = None

        self.path = path
        self.method = method.upper()
        self.environ = environ

    def get_header(self, name, default=None):
        return self.headers.get(name.title(), default)

    @property
    def headers(self):
        if self._headers is not None:
            return self._headers
        self._headers = {}
        for key, value in self.environ.items():
            if key == "CONTENT_TYPE" or key == "CONTENT_LENGTH":
                header = key.title().replace("_", "-")
                self._headers[header] = value
            elif key.startswith("HTTP_"):
                header = key[5:].title().replace("_", "-")
                self._headers[header] = value
        return self._headers

    @property
    def args(self):
        if self._args is not None:
            return self._args
        qs = self.environ["QUERY_STRING"]
        self._args = utils.query_string_to_dict(qs)
        return self._args

    @property
    def json_args(self):
        if self._json_args is not None:
            return self._json_args

        try:
            qs = self.environ["QUERY_STRING"]
            self._json_args = utils.query_string_to_json(qs)
        except Exception:
            raise errors.HTTP_400("Invalid JSON in request query string")

        return self._json_args

    @property
    def body(self):
        if self._body is not None:
            return self._body
        try:
            length = int(self.environ.get("CONTENT_LENGTH", "0"))
        except ValueError:
            length = 0
        if length > 0:
            self._body = self.environ['wsgi.input'].read(length)
        else:
            self._body = ''

        return self._body

    @property
    def cookies(self):
        if self._cookies is not None:
            return self._cookies

        self._cookies = {}

        if 'HTTP_COOKIE' in self.environ:
            for cookie_line in self.environ['HTTP_COOKIE'].split(';'):
                if "=" not in cookie_line:
                    continue  # malformed cookie
                if "DELETED" in cookie_line:
                    continue
                cname, cvalue = cookie_line.strip().split('=', 1)
                self._cookies[cname] = cvalue

        return self._cookies

    @property
    def data(self):
        if self._data is not None:
            return self._data
        if self.body:
            try:
                self._data = json.loads(self.body.decode('utf-8'))
            except Exception:
                raise errors.HTTP_400("Expected JSON in request body")
        else:
            self._data = {}

        return self._data

    @property
    def ip(self):
        try:
            return self.environ['HTTP_X_FORWARDED_FOR'].split(',')[-1].strip()
        except KeyError:
            return self.environ['REMOTE_ADDR']


class Response(object):

    def __init__(self, status_code):
        self.header_dict = {
            "Content-Type": "application/json"
        }
        self._headers = []
        self.cookie_dict = {}
        self.status_code = status_code

    def set_header(self, key, value):
        self.header_dict[key] = value

    def set_cookie(
            self, key, value, expires="", path='/', domain="", flags=[]):
        value = value.replace(";", "")  # ; not allowed
        if expires:
            expires = "expires=%s; " % (expires)
        if domain:
            domain = "Domain=%s; " % (domain)

        self.cookie_dict[key] = "%s;%s%s path=%s; %s"\
            % (value, domain, expires, path, "; ".join(flags))

    def delete_cookie(self, key):
        expiry_string = 'Thu, 01 Jan 1970 00:00:00 GMT'
        self.set_cookie(key, "DELETED", expires=expiry_string)

    @property
    def headers(self):
        self._headers = []
        for k, v in self.header_dict.items():
            self._headers.append((k, v))
        for k, v in self.cookie_dict.items():
            self._headers.append(('Set-Cookie', '%s=%s' % (k, v)))
        return self._headers

    @property
    def status(self):
        if self.status_code in STATUSES:
            return STATUSES[self.status_code]
        print("Warning! Status %s does not exist!" % (self.status_code))
        return STATUSES[577]


class WSGI:

    routes = []
    debug = False
    logger = None
    before = None
    teardown = None
    after = None
    headers = None
    strip_path = True
    json_cls = None

    def __init__(self, environ, start_response):

        if not self.logger:
            self.logger = logging.Logger(__name__)

        self.logger.debug("WSGI __init__ called")

        path = environ["PATH_INFO"]
        if self.strip_path and len(path) > 1 and path.endswith("/"):
            path = path[0:-1]

        self.request = Request(
            path=path,
            method=environ["REQUEST_METHOD"],
            environ=environ
        )

        self.response = Response(status_code=200)

        self.environ = environ
        self.start = start_response
        self.response.set_header("Content-Type", "application/json")

    def __iter__(self):
        try:
            if self.before:
                self.before()
            resp = self.delegate()
            if self.after:
                self.after()
            headers = self.response.headers
            if self.headers:
                headers += self.headers
            self.start(self.response.status, headers)

        except errors.HTTPError as err:
            self.response.status_code = err.status_code
            headers = [("Content-Type", "application/json")]
            if self.headers:
                headers += self.headers
            if err.headers:
                headers += err.headers
            self.start(self.response.status, headers)
            resp = err.response()

        except Exception as err:
            self.logger.exception(err)
            headers = [("Content-Type", "application/json")]
            self.start(STATUSES[500], headers)
            if self.debug:
                resp = {"error": traceback.format_exc().split("\n")}
            else:
                resp = {"error": "Internal server error encountered."}

        if self.teardown:
            self.teardown()

        if isinstance(resp, (dict, list)):
            try:
                if self.debug:
                    jresp = json.dumps(resp, indent=4, cls=self.json_cls)
                else:
                    jresp = json.dumps(resp, cls=self.json_cls)
            except Exception:
                if self.debug:
                    msg = traceback.format_exc().split("\n")
                    jresp = json.dumps({"error": msg}, indent=4)
                else:
                    msg = "An unhandled exception occured during response"
                    jresp = json.dumps({"error": msg})
            self.logger.debug("Sending JSON response: %s", jresp)
            return iter([jresp.encode('utf-8')])
        elif isinstance(resp, str):
            self.logger.debug("Sending string response: %s", resp)
            return iter([resp.encode('utf-8')])
        else:
            self.logger.debug("Sending unknown response: %s", resp)
            return iter(resp)

    def delegate(self):
        path = self.request.path
        method = self.request.method

        for pattern, handler in self.routes:
            # Set defaults for handler
            handler.request = self.request
            handler.response = self.response

            m = re.match('^' + pattern + '$', path)
            if m:
                if hasattr(handler, 'before'):
                    handler.before()

                args = m.groups()
                funcname = method.lower()
                try:
                    func = getattr(handler, funcname)
                except AttributeError:
                    raise errors.HTTP_405("%s not allowed" % (method.upper()))

                output = func(*args)

                if hasattr(handler, 'after'):
                    handler.after()

                return output

        raise errors.HTTP_404("Path %s not found" % (path))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import inspect
import os
import sys

# Local imports
from pycnic.core import WSGI


class DummyFile(object):
    """Class that will no-op writes to stdout and stderr."""
    def write(self, x): pass


class nostdout(object):
    """Capture any stdout and stderr from importing the files.

    During the import of the user-specified files, sometimes code will
    get run. This silences whatever would print during that in order to
    keep the console output of the Pycnic CLI clean.
    """

    def __enter__(self):
        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr
        sys.stdout = DummyFile()
        sys.stderr = DummyFile()
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr


def print_description(cls, max_route_length, max_method_length):
    """Extract the docstring from the class and print it.

    Formatted to fit underneath the class name in all output formats."""
    if not cls.__doc__:
        return

    doc_string = cls.__doc__
    doc_lines = doc_string.split('\n')
    right_justify = max_route_length+2+max_method_length
    print('Description:'.rjust(right_justify) + '  ' + doc_lines[0].strip())
    for line in doc_lines[1:]:
        print(''.rjust(max_route_length+4+max_method_length) + line.strip())
    print('')


def alpha_sort(routes, verbose=False):
    """Sort routes alphabetically by their regex pattern.

    Regexes are not inherently sortable by the patterns they match,
    so this treates the pattern as a string and sorts accordingly.
    """
    routes.sort(key=lambda x: x[0])
    max_route_length = 5  # Length of the word "route"
    max_method_length = 6  # Length of the word "method"
    # Determine justified string lengths
    for route in routes:
        methods_str = ', '.join(route[1])
        max_route_length = max(max_route_length, len(route[0]))
        max_method_length = max(max_method_length, len(methods_str))

    ljust_method_word = 'Method'.ljust(max_method_length)
    ljust_route_word = 'Route'.ljust(max_route_length)
    print(ljust_route_word + '  ' + ljust_method_word + '  Class')
    print('')

    # Print justified strings
    for route in routes:
        ljust_route = route[0].ljust(max_route_length)
        methods_str = ', '.join(route[1]).upper()
        ljust_methods = methods_str.ljust(max_method_length)
        route_cls_name = full_class_name(route[2])
        print('  '.join([ljust_route, ljust_methods, route_cls_name]))
        if verbose:
            print_description(route[2], max_route_length, max_method_length)


def class_sort(routes, verbose=False):
    """Sort routes alphabetically by their class name."""
    routes.sort(key=lambda x: full_class_name(x[2]))
    max_route_length = 5  # Length of the word "route"
    max_method_length = 6  # Length of the word "method"
    # Determine justified string lengths
    for route in routes:
        methods_str = ', '.join(route[1])
        max_route_length = max(max_route_length, len(route[0]))
        max_method_length = max(max_method_length, len(methods_str))

    ljust_method_word = 'Method'.ljust(max_method_length)
    ljust_route_word = 'Route'.ljust(max_route_length)
    print(ljust_route_word + '  ' + ljust_method_word + '  Class')
    print('')

    # Print justified strings
    for route in routes:
        ljust_route = route[0].ljust(max_route_length)
        methods_str = ', '.join(route[1]).upper()
        ljust_methods = methods_str.ljust(max_method_length)
        route_cls_name = full_class_name(route[2])
        print('  '.join([ljust_route, ljust_methods, route_cls_name]))
        if verbose:
            print_description(route[2], max_route_length, max_method_length)


def no_sort(routes, verbose=False):
    """Don't process the list at all, simply format it."""
    max_route_length = 5  # Length of the word "route"
    max_method_length = 6  # Length of the word "method"
    # Determine justified string lengths
    for route in routes:
        methods_str = ', '.join(route[1])
        max_route_length = max(max_route_length, len(route[0]))
        max_method_length = max(max_method_length, len(methods_str))

    ljust_method_word = 'Method'.ljust(max_method_length)
    ljust_route_word = 'Route'.ljust(max_route_length)
    print(ljust_route_word + '  ' + ljust_method_word + '  Class')
    print('')

    # Print justified strings
    for route in routes:
        ljust_route = route[0].ljust(max_route_length)
        methods_str = ', '.join(route[1]).upper()
        ljust_methods = methods_str.ljust(max_method_length)
        route_cls_name = full_class_name(route[2])
        print('  '.join([ljust_route, ljust_methods, route_cls_name]))
        if verbose:
            print_description(route[2], max_route_length, max_method_length)


def method_sort(routes, verbose=False):
    """Group by methods, but do not sort."""
    method_routes = {
        'get': [],
        'head': [],
        'post': [],
        'put': [],
        'delete': [],
        'connect': [],
        'options': [],
        'trace': [],
        'patch': [],
    }

    # Find which routes support which methods, and place into dict accordingly
    for route, methods, route_class in routes:
        for m in methods:
            method_routes[m].append((route, route_class))

    # Start at length of word "METHOD"
    max_method_length = 6  # Length of the word "method"
    for key in method_routes:
        if len(method_routes[key]) > 0:
            max_method_length = max(max_method_length, len(key))

    max_route_length = 5  # Length of the word "route"
    for route in routes:
        max_route_length = max(max_route_length, len(route[0]))

    # Print out the information.
    ljust_method_word = 'Method'.ljust(max_method_length)
    ljust_route_word = 'Route'.ljust(max_route_length)
    print(ljust_method_word + '  ' + ljust_route_word + '  Class')
    for key in method_routes:
        if len(method_routes[key]) <= 0:
            continue
        print('')
        for a in method_routes[key]:
            ljust_route = a[0].ljust(max_route_length)
            ljust_method = key.upper().ljust(max_method_length)
            route_cls_name = full_class_name(a[1])
            print('  '.join([ljust_method, ljust_route, route_cls_name]))
            if verbose:
                print_description(a[1], max_route_length, max_method_length)


def full_class_name(cls):
    """Get the fully qualified class name of an object class.

    Reference: https://stackoverflow.com/a/2020083
    """
    module = cls.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return cls.__class__.__module__  # Avoid reporting __builtin__
    else:
        return module + '.' + cls.__class__.__name__


def find_routes():
    """Find the routes in a specified class.

    This is done by spawning an instance of the class, and inspecting the
    routes attribute of that class.
    """
    # Need to deep copy to preserve original system path
    starting_sys_path = [el for el in sys.path]

    # Parse command-line arguments
    parser = argparse.ArgumentParser(prog='pycnic routes',
                                     description='Pycnic Routes')
    parser.add_argument('-sort', required=False,
                        help='Sorting method to use, if any.',
                        choices=('alpha', 'class', 'method'))
    parser.add_argument('--verbose', '-v',
                        action='store_true',
                        help='Verbose route output. \
                        Prints route class docstring.')
    parser.add_argument('path', type=str,
                        nargs='?', default='main:app',
                        help='Path to the specified application. If none \
                        given, looks in main.py for a class called app. \
                        Example: pycnic.pycnic:app')

    # Since we've already handled the command argument, discard it.
    args = parser.parse_args(sys.argv[2:])

    # Extract class name from path
    module_path, class_name = args.path.split(':')

    # Determine module name, for dynamic import
    module_name = module_path.split('.')[-1]

    # Convert module path-type to os path-type
    file_path = os.path.join(os.getcwd(), *module_path.split('.')) + '.py'

    # Make sure that the file specified actually exists
    if not os.path.isfile(file_path):
        print('File ' + file_path + ' does not exist.')
        exit(1)

    # Add the current working directory to the path so that imports will
    # function properly in the imported file.
    sys.path.append(os.getcwd())

    with nostdout() as _:
        # Dynamically load the file in order to assess the class
        if sys.version_info[0] < 3:
            # For python 2.X, we use the imp library
            # Reference: https://stackoverflow.com/a/67692
            import imp

            loaded_src = imp.load_source(module_name, file_path)
            wsgi_class = getattr(loaded_src, class_name)
        else:
            # For python 3.X, we use importlib
            # Reference: https://stackoverflow.com/a/67692
            if sys.version_info[1] >= 5:
                # For 3.5+, use importlib.util for dynamic module loading
                from importlib.util import spec_from_file_location
                import importlib.util
                spec = spec_from_file_location(module_name, file_path)
                loaded_src = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(loaded_src)
                wsgi_class = getattr(loaded_src, class_name)
            elif sys.version_info[1] in [3, 4]:
                # For 3.3 and 3.4, SourceFileLoader is the tool to use
                from importlib.machinery import SourceFileLoader
                loaded_src = SourceFileLoader(module_name, file_path)
                wsgi_class = getattr(loaded_src, class_name)
            else:
                raise ImportError('Pycnic routes is not supported for Python \
                                  versions 3.0.X through 3.2.X')

    # Restore the system PATH to its original state
    #  now that we've loaded the input project.
    sys.path = starting_sys_path

    # Make sure that the specified element is actually a subclass of the
    # WSGI class provided in pycnic.core
    if not inspect.isclass(wsgi_class):
        print(class_name + ' is not a class in ' + file_path)
        exit(1)
    if not issubclass(wsgi_class, WSGI):
        print(class_name + ' is not a subclass of pycnic.core.WSGI.')
        exit(1)

    # Reference: https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods
    http_methods = [
        'get',
        'head',
        'post',
        'put',
        'delete',
        'connect',
        'options',
        'trace',
        'patch'
    ]

    # Start with empty list of routes
    # This will ultimately be a list of tuples of routes and methods
    all_routes = []

    for route_regex, route_class in wsgi_class.routes:
        # Old-style python2 classes require you to access
        #  object.__class__.__name__.
        class_name = route_class.__class__.__name__

        # Get the intersection of http_methods and class methods to find
        #  what http methods are handled by the class
        class_http_methods = [
            method for method in http_methods if method in dir(route_class)
        ]

        all_routes.append((route_regex, class_http_methods, route_class))

    if args.sort:
        if args.sort == 'method':
            method_sort(all_routes, args.verbose)
        elif args.sort == 'alpha':
            alpha_sort(all_routes, args.verbose)
        elif args.sort == 'class':
            class_sort(all_routes, args.verbose)
    else:
        no_sort(all_routes, args.verbose)


def usage():
    """Print known CLI commands.

    Emulates the style of argparse's output.
    """
    allowed_commands = [
        'routes'
    ]
    allowed_commands_list_str = '{' + ', '.join(allowed_commands) + '}'
    script_name = os.path.basename(sys.argv[0])
    print('usage: ' + script_name + ' [-h] ' + allowed_commands_list_str)
    print('')
    print('Pycnic CLI')
    print('')
    print('positional arguments:')
    print('  ' + allowed_commands_list_str + '      Subcommand to run.')
    print('')
    print('optional arguments:')
    print('  -h, --help  show this help message and exit')


def main():
    """Main function referenced by entry_points in setup.py"""
    if len(sys.argv) < 2:
        usage()
        exit()
    run_type = sys.argv[1]
    if run_type == '-h':
        usage()
    elif run_type.lower() == 'routes':
        find_routes()
    else:
        usage()


if __name__ == "__main__":
    main()



from .data import STATUSES


class PycnicError(Exception):
    pass


class HTTPError(PycnicError):
    status_code = 0
    status = None
    message = None
    data = None
    headers = None

    def __init__(self, status_code, message, data=None, headers=[]):
        if self.status_code:
            status_code = self.status_code
        self.status_code = status_code
        self.status = STATUSES[status_code]
        self.message = message
        self.data = data
        if headers:
            self.headers = headers

    def response(self):
        return {
            "status": self.status,
            "status_code": self.status_code,
            "error": self.message,
            "data": self.data
        }


class HTTPNumeric(HTTPError):

    status_code = 0

    def __init__(self, message, data=None, headers=[]):
        super(HTTPError, self).__init__(self.status_code, message, data, headers)
        self.status = STATUSES[self.status_code]
        self.message = message
        self.data = data
        self.headers = headers


class HTTP_400(HTTPNumeric):
    status_code = 400


class HTTP_401(HTTPNumeric):
    status_code = 401


class HTTP_403(HTTPNumeric):
    status_code = 403


class HTTP_404(HTTPNumeric):
    status_code = 404


class HTTP_405(HTTPNumeric):
    status_code = 405


class HTTP_408(HTTPNumeric):
    status_code = 408


class HTTP_500(HTTPNumeric):
    status_code = 500



#!/usr/bin/env python
from setuptools import setup, find_packages
import os
import sys

os.chdir(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, os.path.dirname(__file__))
from pycnic import __version__

setup(name='pycnic',
    version = __version__,
    description = 'A simple, ultra light-weight, pure-python RESTful JSON API framework.',
    long_description = (
        'Pycnic offers a fully WSGI compliant JSON only web framework for '
        'quickly creating fast, modern web applications '
        'based on AJAX. Static files are served over a CDN or ' 
        'with a standard webserver, like Apache.'),
    author = 'Aaron Meier',
    author_email = 'webgovernor@gmail.com',
    packages = ['pycnic'],
    package_dir={'pycnic':'pycnic'},
    url = 'http://pycnic.nullism.com',
    license = 'MIT',
    entry_points={"console_scripts": ["pycnic = pycnic.cli:main"]},
    install_requires = [],
    provides = ['pycnic']
)




import json
import sys
from functools import wraps

from . import errors

if sys.version_info >= (3, 0):
    from urllib.parse import parse_qsl
else:
    from cgi import parse_qsl


def query_string_to_dict(qs):
    """ Returns a dictionary from a QUERY_STRING """

    pairs = parse_qsl(qs)
    if pairs:
        return dict(pairs)
    return {}


def query_string_to_json(qs):
    """ Returns a dictionary from a QUERY_STRING with JSON in it """

    if "json=" not in qs:
        return {}
    data = query_string_to_dict(qs)
    return json.loads(data.get("json", "{}"))


def requires_validation(validator, with_route_params=False):
    """ Validates an incoming request over given validator.
    If with_route_params is set to True, validator is called with request
    data and args taken from route, otherwise only request data is
    passed to validator. If validator raises any Exception, HTTP_400 is raised.
    """
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            try:
                if with_route_params:
                    validator(args[0].request.data, *args[1:])
                else:
                    validator(args[0].request.data)
            except Exception as e:
                raise errors.HTTP_400(str(e))

            return f(*args, **kwargs)
        return wrapped
    return wrapper
